"""
BioC XML encoder
"""

from lxml import etree

from bioc import BioCCollection, BioCDocument, BioCLocation, BioCNode, BioCRelation, \
    BioCAnnotation, BioCSentence, BioCPassage


def dump(collection: BioCCollection, fp, pretty_print: bool = True):
    """
    Serialize ``collection`` as a BioC formatted stream to ``fp``.

    Args:
        collection: the BioC collection
        fp: a ``.write()``-supporting file-like object
        pretty_print: enables formatted XML
    """
    fp.write(dumps(collection, pretty_print))


def dumps(collection: BioCCollection, pretty_print: bool = True) -> str:
    """
    Serialize ``collection`` to a BioC formatted ``str``.

    Args:
        collection: the BioC collection
        pretty_print: enables formatted XML

    Returns:
        a BioC formatted ``str``
    """
    doc = etree.ElementTree(BioCXMLEncoder().default(collection))
    s = etree.tostring(doc, pretty_print=pretty_print, encoding=collection.encoding,
                       standalone=collection.standalone)
    return s.decode(collection.encoding)


def encode_location(location: BioCLocation):
    """Encode a single location."""
    return etree.Element('location',
                         {'offset': str(location.offset), 'length': str(location.length)})


def encode_node(node: BioCNode):
    """Encode a single node."""
    return etree.Element('node', {'refid': node.refid, 'role': node.role})


def encode_relation(relation: BioCRelation):
    """Encode a single relation."""
    tree = etree.Element('relation', {'id': relation.id})
    encode_infons(tree, relation.infons)
    for node in relation.nodes:
        tree.append(encode_node(node))
    return tree


def encode_infons(tree, infons):
    for k, v in infons.items():
        elem = encode_infon(k, v)
        tree.append(elem)


def encode_infon(k, v):
    elem = etree.Element('infon', {'key': str(k)})
    elem.text = str(v)
    return elem


def encode_annotation(annotation):
    """Encode a single annotation."""
    tree = etree.Element('annotation', {'id': annotation.id})
    encode_infons(tree, annotation.infons)
    for location in annotation.locations:
        tree.append(encode_location(location))
    etree.SubElement(tree, 'text').text = annotation.text
    return tree


def encode_sentence(sentence):
    """Encode a single sentence."""
    tree = etree.Element('sentence')
    encode_infons(tree, sentence.infons)
    etree.SubElement(tree, 'offset').text = str(sentence.offset)
    if sentence.text:
        etree.SubElement(tree, 'text').text = sentence.text
    for ann in sentence.annotations:
        tree.append(encode_annotation(ann))
    for rel in sentence.relations:
        tree.append(encode_relation(rel))
    return tree


def encode_passage(passage):
    """Encode a single passage."""
    tree = etree.Element('passage')
    encode_infons(tree, passage.infons)
    etree.SubElement(tree, 'offset').text = str(passage.offset)
    if passage.text:
        etree.SubElement(tree, 'text').text = passage.text
    for sen in passage.sentences:
        tree.append(encode_sentence(sen))
    for ann in passage.annotations:
        tree.append(encode_annotation(ann))
    for rel in passage.relations:
        tree.append(encode_relation(rel))
    return tree


def encode_document(document):
    """Encode a single document."""
    tree = etree.Element('document')
    etree.SubElement(tree, 'id').text = document.id
    encode_infons(tree, document.infons)
    for passage in document.passages:
        tree.append(encode_passage(passage))
    for ann in document.annotations:
        tree.append(encode_annotation(ann))
    for rel in document.relations:
        tree.append(encode_relation(rel))
    return tree


def encode_collection(collection):
    """Encode a single collection."""
    tree = etree.Element('collection')
    etree.SubElement(tree, 'source').text = collection.source
    etree.SubElement(tree, 'date').text = collection.date
    etree.SubElement(tree, 'key').text = collection.key
    encode_infons(tree, collection.infons)
    for doc in collection.documents:
        tree.append(encode_document(doc))
    return tree


class BioCXMLEncoder:
    """
    Extensible BioC XML encoder for BioC data structures.

    To extend this to recognize other objects, subclass and implement a ``.default()`` method
    with another method that returns an XML tree for ``o`` if possible, otherwise it should
    call the superclass implementation.
    """

    def default(self, obj):
        """Implement this method in a subclass such that it returns a tree for ``o``."""
        if isinstance(obj, BioCLocation):
            return encode_location(obj)
        if isinstance(obj, BioCNode):
            return encode_node(obj)
        if isinstance(obj, BioCRelation):
            return encode_relation(obj)
        if isinstance(obj, BioCAnnotation):
            return encode_annotation(obj)
        if isinstance(obj, BioCSentence):
            return encode_sentence(obj)
        if isinstance(obj, BioCPassage):
            return encode_passage(obj)
        if isinstance(obj, BioCDocument):
            return encode_document(obj)
        if isinstance(obj, BioCCollection):
            return encode_collection(obj)
        raise TypeError(f'Object of type {obj.__class__.__name__} is not BioC XML serializable')

    def encode(self, obj):
        """Encode an obj to an element tree"""
        return self.default(obj)


class BioCXMLDocumentWriter:
    """
    Writer for the BioC XML format, one document at a time.
    """

    def __init__(self, file, encoding='utf8', standalone=True):
        self.encoding = encoding
        self.standalone = standalone
        self.file = file
        self.encoder = BioCXMLEncoder()
        self.__writer = self.__writer__()
        next(self.__writer)  # start writing (run up to 'yield')

    def __writer__(self):
        with etree.xmlfile(self.file, encoding=self.encoding, close=True) as xf:
            xf.write_declaration(standalone=self.standalone)
            with xf.element('collection'):
                try:
                    while True:
                        elem = (yield)
                        xf.write(elem)
                        xf.write('\n')
                        xf.flush()
                except GeneratorExit:
                    pass

    def write_collection_info(self, collection: BioCCollection):
        """
        Writes the collection information: encoding, version, DTD, source, date, key, infons, etc.
        """
        elem = etree.Element('source')
        elem.text = collection.source
        self.__writer.send(elem)

        elem = etree.Element('date')
        elem.text = collection.date
        self.__writer.send(elem)

        elem = etree.Element('key')
        elem.text = collection.key
        self.__writer.send(elem)

        for k, v in collection.infons.items():
            elem = etree.Element('infon', {'key': str(k)})
            elem.text = str(v)
            self.__writer.send(elem)

    def close(self):
        """Close this writer"""
        self.__writer.close()

    def write_document(self, document: BioCDocument):
        """Encode and write a single document."""
        tree = self.encoder.default(document)
        self.__writer.send(tree)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



"""
Validate BioC data structure
"""
from typing import Callable, List

from bioc.bioc import BioCDocument, BioCCollection


# pylint: disable=unused-argument
def _default_error(msg, traceback):
    """Default error handler that accepts two parameters: error message and traceback."""
    raise ValueError(msg)


class BioCValidator():
    """
    Validate BioC data structure
    """

    def __init__(self, onerror: Callable[[str, List], None] = None):
        if onerror is None:
            self.onerror = _default_error
        else:
            self.onerror = onerror
        self.current_docid = None
        self.traceback = []

    def validate_doc(self, document: BioCDocument):
        """Validate a single document."""
        annotations = []
        annotations.extend(document.annotations)
        annotations.extend(document.relations)
        for passage in document.passages:
            annotations.extend(passage.annotations)
            annotations.extend(passage.relations)
            for sentence in passage.sentences:
                annotations.extend(sentence.annotations)
                annotations.extend(sentence.relations)

        self.current_docid = document.id
        self.traceback.append(document)

        text = self.__get_doc_text(document)
        self.__validate_ann(document.annotations, text, 0)
        for relation in document.relations:
            for node in relation.nodes:
                if not self.__contains(annotations, node.refid):
                    self.onerror(f'Cannot find node {node} in document {document.id}',
                                 self.traceback)

        for passage in document.passages:
            self.traceback.append(passage)

            text = self.__get_passage_text(passage)
            self.__validate_ann(passage.annotations, text, passage.offset)
            for relation in passage.relations:
                for node in relation.nodes:
                    if not self.__contains(annotations, node.refid):
                        self.onerror(f'Cannot find node {node} in document {document.id}',
                                     self.traceback)

            for sentence in passage.sentences:
                self.traceback.append(sentence)
                self.__validate_ann(sentence.annotations, sentence.text, sentence.offset)
                for relation in sentence.relations:
                    for node in relation.nodes:
                        if not self.__contains(annotations, node.refid):
                            self.onerror(f'Cannot find node {node} in document {document.id}',
                                         self.traceback)
                self.traceback.pop()
            self.traceback.pop()
        self.traceback.pop()

    def validate(self, collection: BioCCollection):
        """Validate a single collection."""
        for document in collection.documents:
            self.validate_doc(document)

    def __validate_ann(self, annotations, text, offset):
        for ann in annotations:
            self.traceback.append(ann)
            location = ann.total_span
            anntext = text[location.offset - offset: location.offset + location.length - offset]
            if anntext != ann.text:
                self.onerror(
                    '%s: Annotation text is incorrect at %d.\n'
                    '  Annotation: %s\n'
                    '  Actual text: %s' %
                    (self.current_docid, location.offset, anntext, ann.text),
                    self.traceback)
            self.traceback.pop()

    @classmethod
    def __contains(cls, annotations, ann_id):
        for ann in annotations:
            if ann.id == ann_id:
                return True
        return False

    def __fill_newline(self, text, offset):
        dis = offset - len(text)
        if dis < 0:
            self.onerror('%s: Overlap with previous text: len[%d] vs next offset[%d]\n%s'
                         % (self.current_docid, len(text), offset, text),
                         self.traceback)
        if dis > 0:
            text += '\n' * dis
        return text

    def __get_passage_text(self, passage):
        if passage.text:
            return passage.text

        text = ''
        for sentence in passage.sentences:
            self.traceback.append(sentence)
            text = self.__fill_newline(text, sentence.offset)
            if sentence.text:
                text += sentence.text
            else:
                self.onerror(
                    f'{self.current_docid}: BioC sentence has no text: {sentence.offset}',
                    self.traceback)
            self.traceback.pop()
        return text

    def __get_doc_text(self, document):
        text = ''
        for passage in document.passages:
            text = self.__fill_newline(text, passage.offset)
            text += self.__get_passage_text(passage)
        return text


def validate(collection, onerror: Callable[[str, List], None] = None):
    """Validate BioC data structure."""
    BioCValidator(onerror).validate(collection)



"""
Data structures and code to read/write BioC XML.
"""
import time


class InfonsMaxin:
    def __init__(self):
        super(InfonsMaxin, self).__init__()
        self.infons = {}

    def clear_infons(self):
        """
        Clears all information.
        """
        self.infons.clear()


class BioCNode:
    """
    The annotations and/or other relations in the relation.
    """

    def __init__(self, refid: str, role: str):
        """
        Args:
            refid: the id of an annotated object or another relation
            role: the role of how the referenced annotation or other relation participates in
            the current relation
        """
        self.refid = refid
        self.role = role

    def __str__(self):
        return f'BioCNode[refid={self.refid},role={self.role}]'

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.refid == other.refid and self.role == other.role

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash((self.refid, self.role))


class BioCLocation:
    """
    The connection to the original text can be made through the offset and length fields.
    """

    def __init__(self, offset: int, length: int):
        """
        Args:
            offset: the offset of annotation
            length: the length of the annotated text
        """
        self.offset = offset
        self.length = length

    def __str__(self):
        return f'BioCLocation[offset={self.offset},length={self.length}]'

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.offset == other.offset and self.length == other.length

    def __ne__(self, other):
        return not self.__eq__(other)

    def __contains__(self, location):
        if not isinstance(location, BioCLocation):
            raise TypeError(f'Object of type {location.__class__.__name__} is not BioCLocation')
        return self.offset <= location.offset \
               and location.offset + location.length <= self.offset + self.length

    def __hash__(self):
        return hash((self.offset, self.length))

    @property
    def end(self):
        """The end offset of annotation"""
        return self.offset + self.length


class BioCAnnotation(InfonsMaxin):
    """
    Stand off annotation.
    """

    def __init__(self):
        super(BioCAnnotation, self).__init__()
        self.locations = []
        self.id = ''
        self.text = ''

    def add_location(self, location: BioCLocation):
        """
        Adds the location at the specified position in this annotation.

        Args:
            location: the location at the specified position in this annotation
        """
        self.locations.append(location)

    def __str__(self):
        s = 'BioCAnnotation['
        s += 'id=%s,' % self.id
        s += 'text=%s,' % _shorten_text(self.text)
        s += 'infons=[%s],' % ','.join('%s=%s' % (k, v) for (k, v) in self.infons.items())
        s += 'locations=[%s],' % ','.join(str(l) for l in self.locations)
        s += ']'
        return s

    def __repr__(self):
        return str(self)

    @property
    def total_span(self) -> BioCLocation:
        """The total span of this annotation. Discontinued locations will be merged."""
        if not self.locations:
            raise ValueError('BioCAnnotation must have at least one location')
        start = min(l.offset for l in self.locations)
        end = max(l.end for l in self.locations)
        return BioCLocation(start, end - start)

    def __contains__(self, annotation):
        if not isinstance(annotation, BioCAnnotation):
            raise TypeError(f'Object of type {annotation.__class__.__name__} is not BioCAnnotation')
        loc1 = self.total_span
        loc2 = annotation.total_span
        return loc2 in loc1


class BioCRelation(InfonsMaxin):
    """
    Relationship between multiple BioCAnnotations and possibly other BioCRelations
    """

    def __init__(self):
        super(BioCRelation, self).__init__()
        self.id = ''
        self.nodes = []

    def __str__(self):
        s = 'BioCRelation['
        s += 'id=%s,' % self.id
        s += 'infons=[%s],' % ','.join('%s=%s' % (k, v) for (k, v) in self.infons.items())
        s += 'nodes=[%s],' % ','.join(str(n) for n in self.nodes)
        s += ']'
        return s

    def __repr__(self):
        return str(self)

    def add_node(self, node: BioCNode):
        """
        Add the node to this relation

        Args:
            node: node to be added to this relation
        """
        self.nodes.append(node)

    def get_node(self, role: str, default=None) -> BioCNode:
        """
        Get the first node with role

        Args:
            role: role
            default: node returned instead of raising StopIteration

        Returns:
            the first node with role
        """
        return next((node for node in self.nodes if node.role == role), default)


class AnnotationMixin:
    def __init__(self):
        super(AnnotationMixin, self).__init__()
        self.annotations = []
        self.relations = []

    def add_annotation(self, annotation: BioCAnnotation):
        """
        Adds annotation in this sentence.

        Args:
            annotation: the annotation
        """
        self.annotations.append(annotation)

    def clear_annotations(self):
        """
        Clears all annotations.
        """
        del self.annotations[:]

    def clear_relations(self):
        """
        Clears all relations.
        """
        del self.relations[:]

    def add_relation(self, relation: BioCRelation):
        """
        Adds relation in this sentence.

        Args:
            relation: a relation
        """
        self.relations.append(relation)


class BioCSentence(AnnotationMixin, InfonsMaxin):
    """
    One sentence in a {@link BioCPassage}.

    It may contain the original text of the sentence or it might be BioCAnnotations and possibly
    BioCRelations on the text of the passage.

    There is no code to keep those possibilities mutually exclusive. However the currently available
    DTDs only describe the listed possibilities.
    """

    def __init__(self):
        super(BioCSentence, self).__init__()
        self.offset = -1
        self.text = ''

    def __str__(self):
        s = 'BioCSentence['
        s += 'offset=%s,' % self.offset
        s += 'text=%s,' % _shorten_text(self.text)
        s += 'infons=[%s],' % ','.join('%s=%s' % (k, v) for (k, v) in self.infons.items())
        s += 'annotations=[%s],' % ','.join(str(a) for a in self.annotations)
        s += 'relations=[%s],' % ','.join(str(r) for r in self.relations)
        s += ']'
        return s

    def __repr__(self):
        return str(self)


class BioCPassage(AnnotationMixin, InfonsMaxin):
    """
    One passage in a BioCDocument.

    This might be the text in the passage and possibly BioCAnnotations over that text. It could be
    the BioCSentences in the passage. In either case it might include BioCRelations over annotations
    on the passage.
    """

    def __init__(self):
        super(BioCPassage, self).__init__()
        self.offset = -1
        self.text = ''
        self.sentences = list()

    def __str__(self):
        s = 'BioCPassage['
        s += 'offset=%d,' % self.offset
        if self.text is not None:
            s += 'text=%s,' % _shorten_text(self.text)
        s += 'infons=[%s],' % ','.join('%s=%s' % (k, v) for (k, v) in self.infons.items())
        s += 'sentences=[%s],' % ','.join(str(s) for s in self.sentences)
        s += 'annotations=[%s],' % ','.join(str(a) for a in self.annotations)
        s += 'relations=[%s],' % ','.join(str(r) for r in self.relations)
        s += ']'
        return s

    def __repr__(self):
        return str(self)

    def add_sentence(self, sentence: BioCSentence):
        """
        Adds sentence in this passage.

        Args:
            sentence: a sentence
        """
        self.sentences.append(sentence)

    def get_sentence(self, offset: int) -> BioCSentence or None:
        """
        Gets sentence with specified offset

        Args:
            offset: sentence offset

        Return:
            the sentence with specified offset
        """
        for sentence in self.sentences:
            if sentence.offset == offset:
                return sentence
        return None


class BioCDocument(AnnotationMixin, InfonsMaxin):
    """
    One document in the BioCCollection.

    An id, typically from the original corpus, identifies the particular document. It includes
    BioCPassages in the document and possibly BioCRelations over annotations on the document.
    """

    def __init__(self):
        super(BioCDocument, self).__init__()
        self.id = ''
        self.passages = list()

    def __str__(self):
        s = 'BioCDocument['
        s += 'id=%s,' % self.id
        s += 'infons=[%s],' % ','.join('%s=%s' % (k, v) for (k, v) in self.infons.items())
        s += 'passages=[%s],' % ','.join(str(p) for p in self.passages)
        s += 'annotations=[%s],' % ','.join(str(a) for a in self.annotations)
        s += 'relations=[%s],' % ','.join(str(r) for r in self.relations)
        s += ']'
        return s

    def __repr__(self):
        return str(self)

    def add_passage(self, passage: BioCPassage):
        """
        Adds passage in this document.

        Args:
            passage: a passage
        """
        self.passages.append(passage)

    def get_passage(self, offset: int) -> BioCPassage or None:
        """
        Gets passage

        Args:
            offset: passage offset

        Return:
            the passage with specified offset
        """
        for passage in self.passages:
            if passage.offset == offset:
                return passage
        return None


class BioCCollection(InfonsMaxin):
    """
    Collection of documents.

    Collection of documents for a project. They may be an entire corpus or some portion of a corpus.
    Fields are provided to describe the collection.

    Documents may appear empty if doing document at a time IO.
    """

    def __init__(self):
        super(BioCCollection, self).__init__()
        self.encoding = 'utf-8'
        self.version = '1.0'
        self.standalone = True

        self.source = ''
        self.date = time.strftime("%Y-%m-%d")
        self.key = ''

        self.documents = list()

    def add_document(self, document: BioCDocument):
        """
        Adds document in this collection.

        Args:
            document: a document
        """
        self.documents.append(document)

    def __str__(self):
        s = 'BioCCollection['
        s += 'source=%s,' % self.source
        s += 'date=%s,' % self.date
        s += 'key=%s,' % self.key
        s += 'infons=[%s],' % ','.join('%s=%s' % (k, v) for (k, v) in self.infons.items())
        s += 'documents=[%s],' % ','.join(str(d) for d in self.documents)
        s += ']'
        return s

    def __repr__(self):
        return str(self)


def _shorten_text(text: str):
    if len(text) <= 40:
        text = text
    else:
        text = text[:17] + ' ... ' + text[-17:]
    return repr(text)



"""
BioC XML decoder
"""

import io
from typing import TextIO

from lxml import etree

from bioc import BioCCollection, BioCDocument, BioCPassage, BioCSentence, BioCAnnotation, \
    BioCRelation, BioCLocation, BioCNode


class BioCXMLDecoder:
    """
    Reader for the BioC XML format.
    """

    def __init__(self):
        pass

    def decodes(self, s: str) -> BioCCollection:
        """
        Deserialize ``s`` to a BioC collection object.

        Args:
            s: a "str" instance containing a BioC collection

        Returns:
            an object of BioCollection
        """
        tree = etree.parse(io.BytesIO(bytes(s, encoding='UTF-8')))
        collection = self.__parse_collection(tree.getroot())
        collection.encoding = tree.docinfo.encoding
        collection.standalone = tree.docinfo.standalone
        collection.version = tree.docinfo.xml_version
        return collection

    def decode(self, fp: TextIO) -> BioCCollection:
        """
        Deserialize ``fp`` to a BioC collection object.

        Args:
            fp: a ``.read()``-supporting file-like object containing a BioC collection

        Returns:
            an object of BioCollection
        """
        # utf8_parser = etree.XMLParser(encoding='utf-8')
        tree = etree.parse(fp)
        collection = self.__parse_collection(tree.getroot())
        collection.encoding = tree.docinfo.encoding
        collection.standalone = tree.docinfo.standalone
        collection.version = tree.docinfo.xml_version
        return collection

    def __parse_collection(self, tree):
        collection = BioCCollection()
        collection.source = tree.findtext('source')
        collection.date = tree.findtext('date')
        collection.key = tree.findtext('key')
        collection.infons = self.__parse_infons(tree)
        for child in tree.findall('document'):
            collection.add_document(self.__parse_document(child))
        return collection

    def __parse_document(self, tree):
        document = BioCDocument()
        document.id = tree.findtext('id')
        document.infons = self.__parse_infons(tree)
        for child in tree.findall('passage'):
            document.add_passage(self.__parse_passage(child))
        for child in tree.findall('annotation'):
            document.add_annotation(self.__parse_annotation(child))
        for child in tree.findall('relation'):
            document.add_relation(self.__parse_relation(child))
        return document

    def __parse_passage(self, tree):
        passage = BioCPassage()
        passage.offset = int(tree.findtext('offset'))
        passage.infons = self.__parse_infons(tree)
        if tree.find('text') is not None:
            passage.text = tree.findtext('text')
        for child in tree.findall('sentence'):
            passage.add_sentence(self.__parse_sentence(child))
        for child in tree.findall('annotation'):
            passage.add_annotation(self.__parse_annotation(child))
        for child in tree.findall('relation'):
            passage.add_relation(self.__parse_relation(child))
        return passage

    def __parse_sentence(self, tree):
        sentence = BioCSentence()
        sentence.offset = int(tree.findtext('offset'))
        sentence.text = tree.findtext('text')
        sentence.infons = self.__parse_infons(tree)
        for child in tree.findall('annotation'):
            sentence.add_annotation(self.__parse_annotation(child))
        for child in tree.findall('relation'):
            sentence.add_relation(self.__parse_relation(child))
        return sentence

    def __parse_annotation(self, tree):
        annotation = BioCAnnotation()
        annotation.id = tree.attrib['id']
        annotation.infons = self.__parse_infons(tree)
        annotation.text = tree.findtext('text')
        for child in tree.findall('location'):
            annotation.add_location(
                BioCLocation(int(child.attrib['offset']), int(child.attrib['length'])))
        return annotation

    def __parse_relation(self, tree):
        relation = BioCRelation()
        if 'id' in tree.attrib:
            relation.id = tree.attrib['id']
        relation.infons = self.__parse_infons(tree)
        for child in tree.findall('node'):
            relation.add_node(BioCNode(child.attrib['refid'], child.attrib['role']))
        return relation

    @classmethod
    def __parse_infons(cls, tree):
        return {child.attrib['key']: child.text for child in tree.findall('infon')}


class BioCXMLDocumentReader:
    """
    Reader for the BioC XML format, one document per iteration.
    """

    def __init__(self, file):
        if not isinstance(file, str):
            file = str(file)
        self.file = file
        self.__context = iter(etree.iterparse(self.file, events=('start', 'end')))
        self.__state = 0
        self.__event = None
        self.__elem = None
        self.__read()

    def __iter__(self):
        return self

    def __next__(self):
        """
        Reads one BioC document from the XML file.

        Returns:
            BioCDocument: the BioC document
        """
        if self.__document is None:
            raise StopIteration
        else:
            document = self.__document
            self.__read()
            return document

    def __read(self):
        while self.__has_next():
            event, elem = self.__next_event()
            if self.__state == 0:
                if event == 'start':
                    if elem.tag == 'collection':
                        self.__state = 1
                        self.__collection = BioCCollection()
                        # collection information
            elif self.__state == 1:
                if event == 'start':
                    if elem.tag == 'document':
                        self.__document = BioCDocument()
                        self.__state = 2
                elif event == 'end':
                    if elem.tag == 'source':
                        self.__collection.source = elem.text
                    elif elem.tag == 'date':
                        self.__collection.date = elem.text
                    elif elem.tag == 'key':
                        self.__collection.key = elem.text
                    elif elem.tag == 'infon':
                        self.__collection.infons[elem.get('key')] = elem.text
                    elif elem.tag == 'collection':
                        self.__state = 0
                        self.__document = None
                        self.__passage = None
                        self.__sentence = None
            elif self.__state == 2:
                if event == 'start':
                    if elem.tag == 'passage':
                        self.__passage = BioCPassage()
                        self.__state = 3
                    elif elem.tag == 'annotation':
                        self.__document.add_annotation(self.__read_annotation(elem))
                    elif elem.tag == 'relation':
                        self.__document.add_relation(self.__read_relation(elem))
                elif event == 'end':
                    if elem.tag == 'id':
                        self.__document.id = elem.text
                    elif elem.tag == 'infon':
                        self.__document.infons[elem.get('key')] = elem.text
                    elif elem.tag == 'document':
                        self.__state = 1
                        return
            elif self.__state == 3:
                if event == 'start':
                    if elem.tag == 'sentence':
                        self.__sentence = BioCSentence()
                        self.__state = 4
                    elif elem.tag == 'annotation':
                        self.__passage.add_annotation(self.__read_annotation(elem))
                    elif elem.tag == 'relation':
                        self.__passage.add_relation(self.__read_relation(elem))
                elif event == 'end':
                    if elem.tag == 'offset':
                        self.__passage.offset = int(elem.text)
                    elif elem.tag == 'text':
                        self.__passage.text = elem.text
                    elif elem.tag == 'infon':
                        self.__passage.infons[elem.get('key')] = elem.text
                    elif elem.tag == 'passage':
                        self.__state = 2
                        if self.__passage is not None:
                            self.__document.add_passage(self.__passage)
            elif self.__state == 4:
                if event == 'start':
                    if elem.tag == 'annotation':
                        self.__sentence.add_annotation(self.__read_annotation(elem))
                    elif elem.tag == 'relation':
                        self.__sentence.add_relation(self.__read_relation(elem))
                elif event == 'end':
                    if elem.tag == 'offset':
                        self.__sentence.offset = int(elem.text)
                    elif elem.tag == 'text':
                        self.__sentence.text = elem.text
                    elif elem.tag == 'infon':
                        self.__sentence.infons[elem.get('key')] = elem.text
                    elif elem.tag == 'sentence':
                        self.__state = 3
                        if self.__sentence is not None:
                            self.__passage.add_sentence(self.__sentence)

    def __read_annotation(self, start_elem):
        ann = BioCAnnotation()
        ann.id = start_elem.get('id')
        while self.__has_next():
            event, elem = self.__next_event()
            if event == 'start':
                pass
            elif event == 'end':
                if elem.tag == 'text':
                    ann.text = elem.text
                elif elem.tag == 'infon':
                    ann.infons[elem.get('key')] = elem.text
                elif elem.tag == 'location':
                    ann.add_location(BioCLocation(int(elem.get('offset')), int(elem.get('length'))))
                elif elem.tag == 'annotation':
                    return ann
        raise RuntimeError("should not reach here")

    def __read_relation(self, start_elem):
        rel = BioCRelation()
        rel.id = start_elem.get('id')
        while self.__has_next():
            event, elem = self.__next_event()
            if event == 'start':
                pass
            elif event == 'end':
                if elem.tag == 'infon':
                    rel.infons[elem.get('key')] = elem.text
                elif elem.tag == 'node':
                    rel.add_node(BioCNode(elem.get('refid'), elem.get('role')))
                if elem.tag == 'relation':
                    return rel
        raise RuntimeError("should not reach here")

    def __has_next(self):
        try:
            self.__event, self.__elem = next(self.__context)
            return True
        except StopIteration:
            self.__event = None
            self.__elem = None
            return False

    def __next_event(self):
        return self.__event, self.__elem

    def get_collection_info(self):
        """
        Reads the collection information: encoding, version, DTD, source, date, key, infons, etc.

        Returns:
            BioCCollection: the BioC collection that contains only information
        """
        return self.__collection

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        """Close this reader"""
        pass


def load(fp) -> BioCCollection:
    """
    Deserialize ``fp`` to a BioC collection object.

    Args:
        fp: a ``.read()``-supporting file-like object containing a BioC collection

    Returns:
         a object of BioCollection
    """
    return BioCXMLDecoder().decode(fp)


def loads(s: str) -> BioCCollection:
    """
    Deserialize ``s`` to a BioC collection object.

    Args:
        s: a ``str`` instance containing a BioC collection

    Returns:
        an object of BioCollection
    """
    return BioCXMLDecoder().decodes(s)



"""
This module implements a number of iterator building blocks.
"""

from typing import Generator

from bioc import BioCCollection, BioCDocument, BioCPassage, BioCSentence, BioCAnnotation, \
    BioCRelation
from bioc.constants import PASSAGE, DOCUMENT, SENTENCE


def annotations(obj: BioCCollection or BioCDocument or BioCPassage or BioCSentence,
                docid: str = None, level: int = PASSAGE) -> Generator[BioCAnnotation, None, None]:
    """
    Get all annotations in document id.

    Args:
        obj: BioCCollection, BioCDocument, BioCPassage, or BioCSentence
        docid: document id. If None, all documents
        level: one of DOCUMENT, PASSAGE, SENTENCE

    Yields:
        one annotation
    """
    if isinstance(obj, BioCCollection):
        for document in filter(lambda d: docid is None or docid == d.id, obj.documents):
            yield from annotations(document, level=level)
    elif isinstance(obj, BioCDocument):
        if level == DOCUMENT:
            yield from obj.annotations
        elif level in (PASSAGE, SENTENCE):
            for passage in obj.passages:
                yield from annotations(passage, level=level)
        else:
            raise ValueError('level must be DOCUMENT, PASSAGE, or SENTENCE')
    elif isinstance(obj, BioCPassage):
        if level == PASSAGE:
            yield from obj.annotations
        elif level == SENTENCE:
            for sentence in obj.sentences:
                yield from annotations(sentence, level=level)
        else:
            raise ValueError('level must be PASSAGE or SENTENCE')
    elif isinstance(obj, BioCSentence):
        if level == SENTENCE:
            yield from obj.annotations
        else:
            raise ValueError('level must be SENTENCE')
    else:
        raise TypeError(f'Object of type {obj.__class__.__name__} must be BioCCollection, '
                        f'BioCDocument, BioCPassage, or BioCSentence')


def relations(obj: BioCCollection or BioCDocument or BioCPassage or BioCSentence,
              docid: str = None, level: int = PASSAGE) -> Generator[BioCRelation, None, None]:
    """
    Get all relations in document id.

    Args:
        obj: BioCCollection, BioCDocument, BioCPassage, or BioCSentence
        docid: document id. If None, all documents
        level: one of DOCUMENT, PASSAGE, SENTENCE

    Yields:
        one relation
    """
    if isinstance(obj, BioCCollection):
        for document in filter(lambda d: docid is None or docid == d.id, obj.documents):
            yield from relations(document, level=level)
    elif isinstance(obj, BioCDocument):
        if level == DOCUMENT:
            yield from obj.relations
        elif level in (PASSAGE, SENTENCE):
            for passage in obj.passages:
                yield from relations(passage, level=level)
        else:
            raise ValueError('level must be DOCUMENT, PASSAGE, or SENTENCE')
    elif isinstance(obj, BioCPassage):
        if level == PASSAGE:
            yield from obj.relations
        elif level == SENTENCE:
            for sentence in obj.sentences:
                yield from relations(sentence, level=level)
        else:
            raise ValueError('level must be PASSAGE or SENTENCE')
    elif isinstance(obj, BioCSentence):
        if level == SENTENCE:
            yield from obj.relations
        else:
            raise ValueError('level must be SENTENCE')
    else:
        raise TypeError(f'Object of type {obj.__class__.__name__} must be BioCCollection, '
                        f'BioCDocument, BioCPassage, or BioCSentence')


def sentences(obj: BioCCollection or BioCDocument or BioCPassage) \
        -> Generator[BioCSentence, None, None]:
    """
    Get all sentences in document id.

    Args:
        obj: BioCCollection, BioCDocument, or BioCPassage

    Yields:
        one sentence
    """
    if isinstance(obj, BioCCollection):
        for document in obj.documents:
            yield from sentences(document)
    elif isinstance(obj, BioCDocument):
        for passage in obj.passages:
            yield from sentences(passage)
    elif isinstance(obj, BioCPassage):
        yield from obj.sentences
    else:
        raise TypeError(f'Object of type {obj.__class__.__name__} must be BioCCollection, '
                        f'BioCDocument, BioCPassage, or BioCSentence')



"""
BioC JSON decoder
"""

import json

import jsonlines

from bioc.bioc import BioCCollection, BioCSentence, BioCRelation, BioCAnnotation, BioCNode, \
    BioCLocation, BioCPassage, BioCDocument
from bioc.constants import DOCUMENT, PASSAGE, SENTENCE


def parse_collection(obj: dict) -> BioCCollection:
    """Deserialize a dict obj to a BioCCollection object"""
    collection = BioCCollection()
    collection.source = obj['source']
    collection.date = obj['date']
    collection.key = obj['key']
    collection.infons = obj['infons']
    for doc in obj['documents']:
        collection.add_document(parse_doc(doc))
    return collection


def parse_annotation(obj: dict) -> BioCAnnotation:
    """Deserialize a dict obj to a BioCAnnotation object"""
    ann = BioCAnnotation()
    ann.id = obj['id']
    ann.infons = obj['infons']
    ann.text = obj['text']
    for loc in obj['locations']:
        ann.add_location(BioCLocation(loc['offset'], loc['length']))
    return ann


def parse_relation(obj: dict) -> BioCRelation:
    """Deserialize a dict obj to a BioCRelation object"""
    rel = BioCRelation()
    rel.id = obj['id']
    rel.infons = obj['infons']
    for node in obj['nodes']:
        rel.add_node(BioCNode(node['refid'], node['role']))
    return rel


def parse_sentence(obj: dict) -> BioCSentence:
    """Deserialize a dict obj to a BioCSentence object"""
    sentence = BioCSentence()
    sentence.offset = obj['offset']
    sentence.infons = obj['infons']
    sentence.text = obj['text']
    for annotation in obj['annotations']:
        sentence.add_annotation(parse_annotation(annotation))
    for relation in obj['relations']:
        sentence.add_relation(parse_relation(relation))
    return sentence


def parse_passage(obj: dict) -> BioCPassage:
    """Deserialize a dict obj to a BioCPassage object"""
    passage = BioCPassage()
    passage.offset = obj['offset']
    passage.infons = obj['infons']
    if 'text' in obj:
        passage.text = obj['text']
    for sentence in obj['sentences']:
        passage.add_sentence(parse_sentence(sentence))
    for annotation in obj['annotations']:
        passage.add_annotation(parse_annotation(annotation))
    for relation in obj['relations']:
        passage.add_relation(parse_relation(relation))
    return passage


def parse_doc(obj: dict) -> BioCDocument:
    """Deserialize a dict obj to a BioCDocument object"""
    doc = BioCDocument()
    doc.id = obj['id']
    doc.infons = obj['infons']
    for passage in obj['passages']:
        doc.add_passage(parse_passage(passage))
    for annotation in obj['annotations']:
        doc.add_annotation(parse_annotation(annotation))
    for relation in obj['relations']:
        doc.add_relation(parse_relation(relation))
    return doc


def load(fp, **kwargs) -> BioCCollection:
    """
    Deserialize fp (a .read()-supporting text file or binary file containing a JSON document) to
    a BioCCollection object

    Args:
        fp: a file containing a JSON document
        **kwargs:

    Returns:
        BioCCollection: a collection
    """
    obj = json.load(fp, **kwargs)
    return parse_collection(obj)


def loads(s: str, **kwargs) -> BioCCollection:
    """
    Deserialize s (a str, bytes or bytearray instance containing a JSON document) to
    a BioCCollection object.

    Args:
        s(str):
        **kwargs:

    Returns:
        BioCCollection: a collection
    """
    obj = json.loads(s, **kwargs)
    return parse_collection(obj)


class BioCJsonIterReader:
    """
    Reader for the jsonlines format.
    """

    def __init__(self, file: str, level: int):
        if level not in {DOCUMENT, PASSAGE, SENTENCE}:
            raise ValueError(f'Unrecognized level: {level}')

        self.reader = jsonlines.open(file)
        self.reader_iter = iter(self.reader)
        self.level = level

    def __iter__(self):
        return self

    def __next__(self):
        obj = next(self.reader_iter)
        if self.level == DOCUMENT:
            return parse_doc(obj)
        if self.level == PASSAGE:
            return parse_passage(obj)
        if self.level == SENTENCE:
            return parse_sentence(obj)
        raise ValueError

    def close(self):
        """Close this reader"""
        self.reader.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



"""
BioC data structures and encoder/decoder for Python
"""
from .bioc import BioCCollection, BioCDocument, BioCPassage, BioCSentence, BioCAnnotation, \
    BioCRelation, BioCLocation, BioCNode
from .biocitertools import annotations, relations, sentences
from .biocxml import load, loads, BioCXMLDocumentReader, dumps, dump, BioCXMLDocumentWriter
from .constants import PASSAGE, SENTENCE, DOCUMENT
from .utils import get_text, pretty_print
from .validator import validate

__all__ = ['BioCAnnotation', 'BioCCollection', 'BioCDocument', 'BioCLocation', 'BioCNode',
           'BioCPassage', 'BioCRelation', 'BioCSentence', 'load', 'loads', 'dump', 'dumps',
           'validate', 'annotations', 'sentences', 'get_text', 'pretty_print',
           'BioCXMLDocumentWriter', 'BioCXMLDocumentReader', 'biocjson']



"""
BioC JSON encoder
"""

import json

import jsonlines

from bioc.bioc import BioCPassage, BioCNode, BioCAnnotation, BioCLocation, BioCRelation, \
    BioCSentence, BioCCollection, BioCDocument
from bioc.constants import DOCUMENT, PASSAGE, SENTENCE


def dumps(obj, **kwargs) -> str:
    """
    Serialize a BioC ``obj`` to a JSON formatted ``str``.
    """
    return json.dumps(obj, cls=BioCJSONEncoder, **kwargs)


def dump(obj, fp, **kwargs):
    """
    Serialize obj as a JSON formatted stream to fp (a .write()-supporting file-like object)
    """
    return json.dump(obj, fp, cls=BioCJSONEncoder, **kwargs)


class BioCJSONEncoder(json.JSONEncoder):
    """
    Extensible BioC JSON encoder for BioC data structures.
    """

    def default(self, o):
        if isinstance(o, BioCLocation):
            return {
                'offset': o.offset,
                'length': o.length,
            }
        if isinstance(o, BioCAnnotation):
            return {
                'id': o.id,
                'infons': o.infons,
                'text': o.text,
                'locations': [self.default(l) for l in o.locations],
            }
        if isinstance(o, BioCNode):
            return {
                'refid': o.refid,
                'role': o.role,
            }
        if isinstance(o, BioCRelation):
            return {
                'id': o.id,
                'infons': o.infons,
                'nodes': [self.default(n) for n in o.nodes]
            }
        if isinstance(o, BioCSentence):
            return {
                'offset': o.offset,
                'infons': o.infons,
                'text': o.text,
                'annotations': [self.default(a) for a in o.annotations],
                'relations': [self.default(r) for r in o.relations],
            }
        if isinstance(o, BioCPassage):
            return {
                'offset': o.offset,
                'infons': o.infons,
                'text': o.text,
                'sentences': [self.default(s) for s in o.sentences],
                'annotations': [self.default(a) for a in o.annotations],
                'relations': [self.default(r) for r in o.relations],
            }
        if isinstance(o, BioCDocument):
            return {
                'id': o.id,
                'infons': o.infons,
                'passages': [self.default(p) for p in o.passages],
                'annotations': [self.default(a) for a in o.annotations],
                'relations': [self.default(r) for r in o.relations],
            }
        if isinstance(o, BioCCollection):
            return {
                'source': o.source,
                'date': o.date,
                'key': o.key,
                'infons': o.infons,
                'documents': [self.default(d) for d in o.documents],
            }
        return json.JSONEncoder.default(self, o)


class BioCJsonIterWriter:
    """
    Writer for the json lines format.
    """

    def __init__(self, file, level: int):
        if level not in {DOCUMENT, PASSAGE, SENTENCE}:
            raise ValueError(f'Unrecognized level: {level}')

        self.writer = jsonlines.open(file, 'w')
        self.level = level

    def write(self, obj: BioCDocument or BioCPassage or BioCSentence):
        """
        Encode and write a single object.

        Args:
            obj: an instance of BioCDocument, BioCPassage, or BioCSentence

        Returns:

        """
        if self.level == DOCUMENT and not isinstance(obj, BioCDocument):
            raise ValueError
        if self.level == PASSAGE and not isinstance(obj, BioCPassage):
            raise ValueError
        if self.level == SENTENCE and not isinstance(obj, BioCSentence):
            raise ValueError
        self.writer.write(BioCJSONEncoder().default(obj))

    def close(self):
        """Close this writer"""
        self.writer.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



"""
Utilities
"""
from typing import Tuple

from lxml import etree

from bioc import BioCDocument, BioCPassage, BioCSentence


def pad_char(text: str, width: int, char: str = '\n') -> str:
    """Pads a text until length width."""
    dis = width - len(text)
    if dis < 0:
        raise ValueError
    if dis > 0:
        text += char * dis
    return text


def get_text(obj: BioCDocument or BioCPassage or BioCSentence) -> Tuple[int, str]:
    """
    Return text with its offset in the document

    Args:
        obj: BioCDocument, BioCPassage, or BioCSentence

    Returns:
        offset, text
    """
    if isinstance(obj, BioCSentence):
        return obj.offset, obj.text
    if isinstance(obj, BioCPassage):
        if obj.text:
            return obj.offset, obj.text
        text = ''
        for sentence in obj.sentences:
            try:
                text = pad_char(text, sentence.offset - obj.offset, ' ')
                assert sentence.text, f'BioC sentence has no text: {sentence.offset}'
                text += sentence.text
            except:
                raise ValueError(f'Overlapping sentences {sentence.offset}')
        return obj.offset, text
    if isinstance(obj, BioCDocument):
        text = ''
        for passage in obj.passages:
            try:
                text = pad_char(text, passage.offset)
                text += get_text(passage)[1]
            except:
                raise ValueError(f'{obj.id}: overlapping passages {passage.offset}')
        return 0, text
    raise TypeError(f'Object of type {obj.__class__.__name__} must be BioCCollection, '
                    f'BioCDocument, BioCPassage, or BioCSentence')


def pretty_print(source, dest):
    """
    Pretty print the XML file
    """
    parser = etree.XMLParser(remove_blank_text=True)
    if not isinstance(source, str):
        source = str(source)
    tree = etree.parse(source, parser)
    docinfo = tree.docinfo
    with open(dest, 'wb') as fp:
        fp.write(etree.tostring(tree, pretty_print=True,
                                encoding=docinfo.encoding, standalone=docinfo.standalone))



# Always prefer setuptools over distutils
# To use a consistent encoding
from codecs import open
from os import path

from setuptools import setup, find_packages

here = path.abspath(path.dirname((__file__)))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='bioc',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version='1.3.1',

    description='BioC data structures and encoder/decoder for Python',
    long_description=long_description,

    # The project's main homepage.
    url='https://github.com/yfpeng/bioc',

    # Author details
    author='Yifan Peng',
    author_email='yifan.peng@nih.gov',

    license='BSD 3-Clause License',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 4 - Beta',
        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: BSD License',

        'Operating System :: MacOS',
        'Operating System :: POSIX',
        'Operating System :: POSIX :: Linux',
        'Operating System :: Microsoft :: Windows',

        # Specify the Python versions you support here.
        'Programming Language :: Python',
        'Topic :: Text Processing :: Markup :: XML',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
    ],

    keywords='bioc',

    packages=find_packages(exclude=["tests.*", "tests"]),
    install_requires=[
        'docutils==0.14',
        'lxml==4.2.5',
        'jsonlines==1.2.0'],
)



"""
Deprecated. Please use `bioc.biocxml.encoder`
"""

import warnings

import bioc.biocxml.encoder


def encode_document(obj):
    """Encode a single document."""
    warnings.warn("deprecated. Please use bioc.biocxml.encoder.encode_document", DeprecationWarning)
    return bioc.biocxml.encoder.encode_document(obj)


def encode_passage(obj):
    """Encode a single passage."""
    warnings.warn("deprecated. Please use bioc.biocxml.encoder.encode_passage", DeprecationWarning)
    return bioc.biocxml.encoder.encode_passage(obj)


def encode_sentence(obj):
    """Encode a single sentence."""
    warnings.warn("deprecated. Please use bioc.biocxml.encoder.encode_sentence", DeprecationWarning)
    return bioc.biocxml.encoder.encode_sentence(obj)


def encode_annotation(obj):
    """Encode a single annotation."""
    warnings.warn("deprecated. Please use bioc.biocxml.encoder.encode_annotation",
                  DeprecationWarning)
    return bioc.biocxml.encoder.encode_annotation(obj)


def encode_relation(obj):
    """Encode a single relation."""
    warnings.warn("deprecated. Please use bioc.biocxml.encoder.encode_relation", DeprecationWarning)
    return bioc.biocxml.encoder.encode_relation(obj)


def encode_infon(k, v):
    warnings.warn("deprecated. Please use bioc.biocxml.encoder.encode_infon", DeprecationWarning)
    return bioc.biocxml.encoder.encode_infon(k, v)



"""
BioC JSON and JSON lines encoder and decoder
"""
from bioc.biocjson.decoder import loads, load, BioCJsonIterReader
from bioc.biocjson.encoder import dumps, dump, BioCJsonIterWriter

__all__ = ['BioCJsonIterReader', 'BioCJsonIterWriter', 'dumps', 'dump', 'load', 'loads']
